


# This proc is only a hack to simplify the setup.
# Many parameters are hardcoded here. Script writer
# should come up with his/her own preparation script.
proc demo_prep {mode} {
    mpstress_prep $mode

    if {$mode == "npar"} {
        map_ring2chnl 0 16
        map_ring2chnl 1 17
        map_ring2chnl 2 18
        map_ring2chnl 3 19
        set_ring_info 0 3 wfq 1 1500 1
        set_ring_info 1 3 wfq 1 3000 1
        set_ring_info 2 3 wfq 1 4500 1
        set_ring_info 3 3 wfq 1 6000 1
    }
}



proc cpusanity {{level 3}} {
    cpu -cpu chimp -no_grc
    cpu -cpu tang -no_grc
}



proc l2sanity {{level 3}} {
    # This needs to run in some loopback mode.
    
    if {$level > 3} { 
        error "usage: l2sanity <level>; where <level> is 0, 1, 2, or 3"
    }

    # Sanity L2 test 1, bursts of packets, 1BD/pkt
    l2spec gen spec
    l2spec dump $spec orig_spec

    if {$level > 0} {
        set size_list [list 64 120 512 1100 1514]
        set duration1 20
        set duration2 10
    } else {
        set size_list [list 64 512 1514]
        set duration1 5
        set duration2 3
    }
    # Send a burst of 50 pkts for $duration1 seconds in poll mode
    l2spec update $spec [list SN auto:4096 pktCnt 50]
    foreach size $size_list {
        puts "**** Packet size: [expr $size + 4] bytes, $duration1-sec bursts of 50 pkts"
        l2spec update $spec [list maxPktLen [expr $size + 4]]
        l2pkt gen $spec pkt_list
        l2pkt_split $pkt_list {0}
        # First arg is a list of packet lists, not just a single packet list.
        l2pkt_loop [list $pkt_list] current -duration $duration1 -poll -skip_warn -dbg
        delay ms 1000
    }
    # Send a burst of 90 pkts for $duration2 seconds in poll mode
    l2spec update $spec [list SN auto:4096 pktCnt 90]
    foreach size $size_list {
        puts "**** Packet size: [expr $size + 4] bytes, $duration2-sec bursts of 90 pkts"
        l2spec update $spec [list maxPktLen [expr $size + 4]]
        l2pkt gen $spec pkt_list
        l2pkt_split $pkt_list {0}
        # First arg is a list of packet lists, not just a single packet list.
        l2pkt_loop [list $pkt_list] current -duration $duration2 -poll -skip_warn -dbg
    }

    # Send a burst of 50 pkts for $duration1 seconds in interrupt mode
    l2spec update $spec [list SN auto:4096 pktCnt 50]
    lmdev set_intr 1
    foreach size $size_list {
        puts "**** Packet size: [expr $size + 4] bytes, $duration1-sec bursts of 50 pkts"
        l2spec update $spec [list maxPktLen [expr $size + 4]]
        l2pkt gen $spec pkt_list
        l2pkt_split $pkt_list {0}
        # First arg is a list of packet lists, not just a single packet list.
        l2pkt_loop [list $pkt_list] current -duration $duration1 -skip_warn -dbg
        delay ms 1000
    }
    lmdev set_intr 0

    # Sanity L2 test 2, l2lb case 3 (multiple BDs)
    if {$level > 0} {
        l2lb 3 all
    } else {
        l2lb 3 2
        l2lb 3 8
        l2lb 3 16
        l2lb 3 31
    }

}

proc loop {iterations args} {
    if {$iterations <= 0} {
        error "Loop: Iterations ($itersions) <= 0"
    }
    for {set idx 0} {$idx < $iterations} {incr idx} {
        eval $args
    }
}

proc quick_test {{test_list "all"} {logfile_suffix "date"}} {
    set full_test_list {
        { reg       "regtest -cof "} \
        { grc       "grc -cof "} \
        { cpuh      "cpuhost -cof"} \
        { mem       "memtest -cof "} \
        { msix      "msixsan "} \
        { l2s       "l2traffic standard poll brief"} \
        \
        { l2j       "l2traffic jumbo poll brief"} \
        { cso       "cso -lb XMAC -cof -mbd {31} -case {1-100} "} \
        { lso       "lso -lb XMAC -cof -mbd {31} -case {1-100} "} \
        { rss       "rss -lb XMAC -cof -case {1-100} "} \
        { hds       "hds -lb XMAC -cof -case {1-100} "} \
        { l2jp      "l2traffic jumbo_aggr poll brief"} \
        { lb        "lb -lb XMAC -cof -case {1-100} "} \
        { cos       "cos -lb XMAC -cof -case {1-100} "} \
        { wol       "wol -cof -case {1-100}"} \
        { tpa       "tpa -lb XMAC -cof -case {1-100} "} \
        { tpac      "tpa -lb XMAC -concur -cof -case {1-100} "} \
        \
        { vfl2      "device vf_en; cso -lb XMAC -mbd 1 -unicast all -spf 1 -epf 1 -svf 0 -evf 8 -poll -cof -case {1-100} "} \
        { csov      "cso -lb XMAC -virt -cof -mbd {31} -case {1-100} "} \
        { lsov      "lso -lb XMAC -virt -cof -mbd {31} -case {1-100} "} \
        { lsv1      "lso -lb XMAC -virt -cof -mbd {31} -case {1-100} -tlist 1"} \
        { lsv2      "lso -lb XMAC -virt -cof -mbd {31} -case {1-100} -tlist 2"} \
        { lsv3      "lso -lb XMAC -virt -cof -mbd {31} -case {1-100} -tlist 3"} \
        { lsv4      "lso -lb XMAC -virt -cof -mbd {31} -case {1-100} -tlist 4"} \
        { cosv      "cos -lb XMAC -virt -cof -case {1-100} "} \
        { hdsv      "hds -lb XMAC -virt -cof -case {1-100} "} \
        { lbv       "lb -lb XMAC -virt -cof -case {1-100} "} \
        { tpav      "tpa -lb XMAC -virt -cof -case {1-100} "} \
        \
        { cfaa      "cfa_ar -lb XMAC -cof -case {1-100} "} \
        { cft0      "cfa_tcam -lb XMAC -palist 0 -cof -case {1-100} "} \
        { cft1      "cfa_tcam -lb XMAC -palist 1 -cof -case {1-100} "} \
        { cft2      "cfa_tcam -lb XMAC -palist 2 -cof -case {1-100} "} \
        { cft3      "cfa_tcam -lb XMAC -palist 3 -cof -case {1-100} "} \
        { cfav      "cfa_veb -lb XMAC -cof -case {1-100} "} \
        { cfac      "cfa_counter -lb XMAC -cof -case {1-100} "} \
        { cfl0      "cfa_lk -lb XMAC -emlist {0-5} -cof -case {1-100} "} \
        { cfl1      "cfa_lk -lb XMAC -emlist {6-10} -cof -case {1-100} "} \
        { cfl2      "cfa_lk -lb XMAC -emlist {11-15} -cof -case {1-100} "} \
        { cfl3      "cfa_lk -lb XMAC -emlist {16-20} -cof -case {1-100} "} \
        { cfl4      "cfa_lk -lb XMAC -emlist {21-25} -cof -case {1-100} "} \
        { cfae      "cfa_eem -lb XMAC -cof -case {1-100} "} \
        { cfm       "cfa_mcr_tx -lb XMAC -cof -case {1-100} "} \
        { cfam      "cfa_mcr -lb XMAC -cof -case {1-100} "} \
        { cfaq      "cfa_qcn -lb XMAC -cof -case {1-100} "} \
        \
        { tdis      "tdi -cof -std -case {1 666 1040} "} \
        { tdij      "tdi -cof -jumbo -case {1 666 1040} "} \
        { tdec      "tedec -cof -case {1 201 441} "} \
        { pfc       "pfc -cof -case {1-5} "} \
        \
        { npl2      "cso -lb XMAC -mbd 1 -unicast all -spf 1 -epf 2 -svf 0 -evf 0 -poll -cof -case {1-100} "} \
        \
        { chmp      "cpu -cpu chimp -cof -case {1-5 10-12 20 36-37 44 59 60-64 89-92 113-115} "} \
        { grcp      "grcp -cof -case {1-100} "} \
        { ape       "cpu -cpu ape -cof -case {1-4 7 10 14-15 20} "} \
        { kcs_on    "kcs -enable "} \
        { kcs       "kcs -test "} \
        { kcs_off   "kcs -disable "} \
        { urd_on    "urd -enable "} \
        { urd       "urd -on; urd -test; urd -off "} \
        { urd_off   "urd -disable "} \
        { tang      "cpu -cpu tang -cof -case {1-4 7 10 14-15 22-24} "} \
        { kong      "cpu -cpu kong -cof -case {1-4 7 10 13 19-21 26-27 32-34 43-44 47-48 67-68 78-79} "} \
        \
        { stat      "stat -poll -lb XMAC -cof -case {1-40} "} \
        { mtf       "mtf -lb XMAC -cof -case {1-100} "} \
    }
    foreach item $full_test_list {
        lappend name_list [lindex $item 0]
        lappend cmd_list [lindex $item 1]
    }

    # Establish $exec_test_list
    if {$test_list == "all"} {
        set exec_test_list $name_list
    } elseif {$test_list == "-help"} {
        puts "Available items are: "
        set idx 0
        foreach item $name_list {
            puts -nonewline [format "%-4s " $item]
            incr idx
            if {($idx % 12 == 0) || ([llength $name_list] == $idx)} {
                puts ""
            }
        }
        return
    } else {
        # Check if user enters a valid list of tests
        foreach item $test_list {
            if {[lsearch -exact $name_list $item] == -1} {
                error "Invalid test item: $item."
            }
        }
        set exec_test_list $test_list
    }
    # Establish log suffix
    if {$logfile_suffix == "date"} {
        set suffix [clock format [clock seconds] -format %m%d]
    } elseif {[string length $logfile_suffix] > 4} {
        error "Log file suffix must be 4 letters or less."
    } else {
        set suffix $logfile_suffix
    }

    set reset_all "reset all"
    # Build the command and run it.
    foreach testmodule $exec_test_list {
        set logfile ""
        append logfile $testmodule $suffix ".log"
        set idx [lsearch -exact $name_list $testmodule]
        set testcmd [lindex $cmd_list $idx]
        # Skip
        switch -- $testmodule {
            "kcs_on"    -
            "kcs_off"   -
            "urd_on"    -
            "urd_off"   {eval "$reset_all; $testcmd"}
            "kcs"       -
            "urd"       -
            "msix"      -
            "l2s"       -
            "l2j"       -
            "l2jp"      {eval "$reset_all; log open $logfile; $testcmd; log close"}
            default     {eval "$reset_all; $testcmd -log $logfile "}
        }
        #eval "$testcmd -log $logfile "
    }
}


# logfile_suffix: suffix of the log files, but be less than 4 letters.
# test_list: a list of tests to execute
proc regression_test {{test_list "all"} {logfile_suffix "date"}} {
    # List the tests, must only be 4 letters or less.
    # Exceptions: kcs_on, kcs_off, urd_on, urd_off as they are not 
    # logged.
    if {$::current(CHIP_CFG) == "ASIC"} {
        set ape_opt "-smbus -i2c "
    } else {
        set ape_opt ""
    }
    set full_test_list [ list \
        { reg       "regtest -cof "} \
        { grc       "grc -cof "} \
        { cpuh      "cpuhost -cof"} \
        { mem       "memtest -cof "} \
        { msix      "msixsan "} \
        { l2s       "l2traffic standard poll "} \
        \
        { l2j       "l2traffic jumbo poll "} \
        { cso       "cso -lb XMAC -cof -mbd {4 8 16 31} "} \
        { lso       "lso -lb XMAC -cof -mbd {4 8 16 31} "} \
        { rss       "rss -lb XMAC -cof "} \
        { hds       "hds -lb XMAC -cof "} \
        { l2jp      "l2traffic jumbo_aggr poll "} \
        { lb        "lb -lb XMAC -cof "} \
        { cos       "cos -lb XMAC -cof "} \
        { wol       "wol -cof "} \
        { tpa       "tpa -lb XMAC -cof "} \
        { tpac      "tpa -lb XMAC -concur -cof "} \
        \
        { vfl2      "device vf_en; cso -lb XMAC -mbd 1 -unicast all -spf 1 -epf 1 -svf 0 -evf 8 -poll -cof "} \
        { csov      "cso -lb XMAC -virt -cof -mbd {4 8 16 31} "} \
        { lsov      "lso -lb XMAC -virt -cof -mbd {4 8 16 31} "} \
        { lsv1      "lso -lb XMAC -virt -cof -mbd {4 8 16 31} -tlist 1"} \
        { lsv2      "lso -lb XMAC -virt -cof -mbd {4 8 16 31} -tlist 2"} \
        { lsv3      "lso -lb XMAC -virt -cof -mbd {4 8 16 31} -tlist 3"} \
        { lsv4      "lso -lb XMAC -virt -cof -mbd {4 8 16 31} -tlist 4"} \
        { lsva      "lso -lb XMAC -virt -cof -mbd {4 8 16 31} -case {1-195010}"} \
        { lsvb      "lso -lb XMAC -virt -cof -mbd {4 8 16 31} -case {195000-end}"} \
        { cosv      "cos -lb XMAC -virt -cof "} \
        { hdsv      "hds -lb XMAC -virt -cof "} \
        { lbv       "lb -lb XMAC -virt -cof "} \
        { tpav      "tpa -lb XMAC -virt -cof "} \
        \
        { cfaa      "cfa_ar -lb XMAC -cof "} \
        { cft0      "cfa_tcam -lb XMAC -palist 0 -cof "} \
        { cft1      "cfa_tcam -lb XMAC -palist 1 -cof "} \
        { cft2      "cfa_tcam -lb XMAC -palist 2 -cof "} \
        { cft3      "cfa_tcam -lb XMAC -palist 3 -cof "} \
        { cfav      "cfa_veb -lb XMAC -cof "} \
        { cfac      "cfa_counter -lb XMAC -cof "} \
        { cfl0      "cfa_lk -lb XMAC -emlist {0-5} -cof  "} \
        { cfl1      "cfa_lk -lb XMAC -emlist {6-10} -cof  "} \
        { cfl2      "cfa_lk -lb XMAC -emlist {11-15} -cof  "} \
        { cfl3      "cfa_lk -lb XMAC -emlist {16-20} -cof  "} \
        { cfl4      "cfa_lk -lb XMAC -emlist {21-25} -cof  "} \
        { cfae      "cfa_eem -lb XMAC -cof "} \
        { cfm       "cfa_mcr_tx -lb XMAC -cof "} \
        { cfam      "cfa_mcr -lb XMAC -cof "} \
        { cfaq      "cfa_qcn -lb XMAC -cof "} \
        \
        { tdis      "tdi -cof -std "} \
        { tdij      "tdi -cof -jumbo "} \
        { tdec      "tedec -cof "} \
        { pfc       "pfc -cof "} \
        \
        { npl2      "cso -lb XMAC -mbd 1 -unicast all -spf 1 -epf 2 -svf 0 -evf 0 -poll -cof "} \
        \
        { chmp      "cpu -cpu chimp -level 6 -cof "} \
        { bono      "cpu -cpu bono -level 6 -cof "} \
        { grcp      "grcp -cof "} \
        [list ape       "cpu -cpu ape -level 6 $ape_opt -cof "] \
        { kcs_on    "kcs -enable "} \
        { kcs       "kcs -test "} \
        { kcs_off   "kcs -disable "} \
        { urd_on    "urd -enable "} \
        { urd       "urd -on; urd -test; urd -off "} \
        { urd_off   "urd -disable "} \
        { tang      "cpu -cpu tang -level 6 -cof "} \
        { kong      "cpu -cpu kong -level 6 -cof "} \
        \
        { stat      "stat -lb XMAC -cof "} \
        { mtf       "mtf -lb XMAC -cof "} \
        { eprm       "eprm"} \
        { push       "push -cof"} \
    ]
    foreach item $full_test_list {
        lappend name_list [lindex $item 0]
        lappend cmd_list [lindex $item 1]
    }

    # Establish $exec_test_list
    if {$test_list == "all"} {
        set exec_test_list $name_list
    } elseif {$test_list == "-help"} {
        puts "Available items are: "
        set idx 0
        foreach item $name_list {
            puts -nonewline [format "%-4s " $item]
            incr idx
            if {($idx % 12 == 0) || ([llength $name_list] == $idx)} {
                puts ""
            }
        }
        return
    } else {
        # Check if user enters a valid list of tests
        foreach item $test_list {
            if {[lsearch -exact $name_list $item] == -1} {
                error "Invalid test item: $item."
            }
        }
        set exec_test_list $test_list
    }
    # Establish log suffix
    if {$logfile_suffix == "date"} {
        set suffix [clock format [clock seconds] -format %m%d]
    } elseif {[string length $logfile_suffix] > 4} {
        error "Log file suffix must be 4 letters or less."
    } else {
        set suffix $logfile_suffix
    }

    set reset_all "reset all"
    # Build the command and run it.
    foreach testmodule $exec_test_list {
        set logfile ""
        append logfile $testmodule $suffix ".log"
        set idx [lsearch -exact $name_list $testmodule]
        set testcmd [lindex $cmd_list $idx]
        # Skip
        switch -- $testmodule {
            "kcs_on"    -
            "kcs_off"   -
            "urd_on"    -
            "urd_off"   {eval "$reset_all; $testcmd"}
            "kcs"       -
            "urd"       -
            "msix"      -
            "l2s"       -
            "l2j"       -
            "l2jp"      {eval "$reset_all; log open $logfile; $testcmd; log close"}
            default     {eval "$reset_all; $testcmd -log $logfile "}
            #default     {puts "$reset_all; $testcmd -log $logfile "}
        }
        #eval "$testcmd -log $logfile "
    }
}


# Returns a list of prod, cons, and cmp (if tx|rx) indices
proc ring_dump {type idx} {
    # Tx, Rx, Cmp ring index offset
    set ring_offset_list [list {$::BDETBD(bd_producer_idx) \
                                $::BDETBD(bd_req_consumer_idx) \
                                $::BDETBD(bd_cmpl_consumer_idx)} \
                               {$::BDERBD(bd_producer_idx) \
                                $::BDERBD(bd_req_consumer_idx) \
                                $::BDERBD(bd_cmpl_consumer_idx)} \
                               {$::CAG(producer_index) $::CAG(consumer_index) 0xffffffff}]
    switch -exact -- $type {
        "tx" {set item 0}
        "rx" {set item 1}
        "cmp" {set item 2}
        default {error "Unknown ring type $type, sb 'tx', 'rx', or 'cmp'"}
    }
    set offsets [lindex $ring_offset_list $item]
    set prod_offset [expr [lindex $offsets 0] + $idx * 4]
    set cons_offset [expr [lindex $offsets 1] + $idx * 4]
    set cmp_offset [expr [lindex $offsets 2] + $idx * 4]
    set result {}
    lappend result [reg read $prod_offset]
    lappend result [reg read $cons_offset]
    if {[lindex $offsets 2] != 0xffffffff} {
        lappend result [reg read $cmp_offset]
    } else {
        # Return a dummy field.
        lappend result 0xffffffff
    }
    return $result
}


proc reg_read_test {start byte_cnt {logfile ""}} {
    if {$logfile != ""} {
        log open $logfile
    }
    # Add any register(s) that should be skipped. Need to be in
    # an increasing order (e.g. [list 0x64320 0x64324]).
    set skip_list [list ]
    set skip_idx 0
    set end [expr $start + $byte_cnt]
    for {set idx $start} {$idx < $end} {incr idx 4} {
        if {$idx == [lindex $skip_list $skip_idx]} {
            incr skip_idx
            continue
        }
        puts "Offset [hex32 $idx] twice"
        reg read $idx
        reg read $idx
    }
    if {$logfile != ""} {
        log close
    }
}

proc mtx_test_init {ring_num } {
    lmdev set_ring tx $ring_num
        set  ::current(TX_RING_TYPE)  RR
        driver load

}
proc mtx_test { pktCnt} {
        set rxPktList {}
        set timeoutCount 300

        
        l2spec gen x
        l2spec dump $x z
# Append serial number in last 4 bytes of payload, starts at 0x1000
# Establish a default spec in $::z
        set ::z(SN) auto:4096
        set ::z(maxPktLen) 68

        # Generate a packet
        l2spec update $x z
        l2pkt gen $x y
        l2pkt_split $y {0}
        l2pkt send $y $pktCnt

        # Now, We need to use driver poll to watch packets in FPGA.
        if {$::current(CHIP_BUILD) == "FPGA"} {
                delay ms 2000
                driver poll
                                            
        }
                                        
        # Wait for packets to be received (till timeout)
        for {set xx 0} {$xx < $timeoutCount} {incr xx} {
        #puts "Waiting for packets ($j out of $timeoutCount)"
                set tmpList [l2pkt receive $pktCnt]

                # Create the received packets list
                for {set yy 0} {$yy < [llength $tmpList]} {incr yy} {
                        lappend rxPktList [lindex $tmpList $yy]
                }
                set myRxPkt [lindex $rxPktList 0]
                if {[llength $rxPktList] == $pktCnt} {
                        #puts "Received all packets."
                        break
                }
                        delay ms 10
        }
        if {$x == $timeoutCount} {
                #puts "ERROR: Time out in receiving packets."
                set errorMsg "ERROR: Time out in receiving packets."
                incr errorCount
                                            
        }
        if {[llength $rxPktList] != $pktCnt} {
                puts "Fail: Only receive [llength $rxPktList], it should be $pktcnt packets."
        } else {
                puts "Received all packets."

        }
        

}



proc wol_test { rule mac } {
#Enable RXP_REG_WOL_MAGICPKT_ENABLE
    set rxp_base 0x60000
    set wol_enable_base 0x5548
    set wol_mag_mac_lsb_base 0x57b0
    set wol_mag_mac_msb_base 0x57b4
    set wol_mag_enable_base 0x5568
    
    set enable_offset [expr $rxp_base + $wol_mag_enable_base + [expr $rule * 0x10]]
    set mac_lsb_offset [expr $rxp_base + $wol_mag_mac_lsb_base + [expr $rule * 0x68]]
    set mac_msb_offset [expr $rxp_base + $wol_mag_mac_msb_base + [expr $rule * 0x68]]

    
    #puts "enable_offset $enable_offset"

        # Enable rule filter
    reg write $enable_offset 0
    

    set mac6 [string range $mac 0 1]
    set mac5 [string range $mac 3 4]
    set mac4 [string range $mac 6 7]
    set mac3 [string range $mac 9 10]
    set mac2 [string range $mac 12 13]
    set mac1 [string range $mac 15 16]

        #puts "Mac addr $mac1 $mac2 $mac3 $mac4 $mac5 $mac6"


        # set rule mac address
        # puts "Mac Address [expr [expr 0x$mac1 ] + [expr 0x$mac2 << 8] + [expr 0x$mac2 << 16] + [expr 0x$mac4 << 24]]"
        reg write $mac_lsb_offset [expr [expr 0x$mac1 ] + [expr 0x$mac2 << 8] + [expr 0x$mac3 << 16] + [expr 0x$mac4 << 24]]
        reg write $mac_msb_offset [expr [expr 0x$mac5 ] + [expr 0x$mac6 << 8]]


        # set Enable Bit
        reg write [expr $rxp_base + $wol_enable_base ] 1
        
}

proc wol_test_show {} {

        reg show 0x65558 0x10

}


proc vlanpkt {cos} {

l2spec gen pkt
l2spec update $pkt [list l3type ip vlanMode specified vlanTag 0xff]
l2spec update $pkt [list vlanhwinsert software vlanQoSVal $cos maxpktlen 100]
l2pkt gen $pkt txpkt
l2pkt send $txpkt

}

proc hw_vlan {vlanid} {

    if {  ($::current(DRV_STATE) != "RUNNING")}  { 
        lmdev vlan_remove 1
        driver load 
    }


    l2spec gen pkt
    l2spec update $pkt [list l3type ip vlanMode specified vlanTag $vlanid]
    l2spec update $pkt [list vlanhwinsert hardware  maxpktlen 100]
    l2pkt gen $pkt txpkt
    l2pkt send $txpkt


}


proc sendrpkt {file} {
    l2spec gen  pkt
    l2spec  update  $pkt  [list  pktlog   nvgre.raw]
    l2spec  update  $pkt  [list  ipChksumGen hardware  tcpChksumGen hardware tunnelIpChksumGen hardware]
    l2pkt gen $pkt  txpkt
    l2pkt send  $txpkt

}


proc cso_all {log} {
    set pfNum 1
    set vfNum 4
    set id 0

    for {set x 1} {$x <= $pfNum} {incr x} {
        device $x
        if {$x == 1} {
            device vf_en
        }
        driver load
        for {set y 1} {$y <= $vfNum} {incr y} {
            if {$x > 1} {
                continue
            }
            device $x.$y
            driver load
            

        }
        
    }

    for {set x 1} {$x <= $pfNum} {incr x} {
        for {set y 0} {$y <= $vfNum} {incr y} {
            if {($x != 1) && ($y > 0)} {
                continue
            }
            device $x.$y


                for {set m 1} {$m <= $pfNum} {incr m} {
                    for {set n 0} {$n < $vfNum} {incr n} {
                        if {($m != 1) && ($n > 0)} {
                            continue
                        }
                        set csvFile "$log\_$id.csv"

                        csotest -no_driver_load -mBD 2 -unicast $m.$n -poll -csv $csvFile
                        incr id

                    }


                    }
            
            
        

        }
    }

}

# This programs the Tx BDs with the same pkt and writes to
# the doorbell. The max number of packets is 500.
# It is assumed to have a single BD page (no page table).
# It is assumed to use 32-bit host address.
proc burst_tx {{ring 0} {pkt_cnt 100} {pkt_sz 64}} {
    load_drv -lb XMAC_LB
    # Generate a min-size packet
    l2spec gen spec
    l2spec update $spec [list maxPktLen $pkt_sz]
    l2pkt gen $spec pkt
    # Since HW generates CRC, the actual pkt length is 4 bytes less.
    incr pkt_sz -4
    # Pretend this is broken into two chunks: 0x20 bytes in first chunk.
    l2pkt_split $pkt {0}
    set pkt_addr0 [lindex [l2pkt virt_addr [lindex $pkt 0]] 0]
    set pkt_addr1 [expr $pkt_addr0 + 0x20]
    # Allocate a chunk of memory with 0x40 byte alignment.
    set addr [hmem alloc [expr $pkt_sz + 0x40]]
    set addr_lo [lindex $addr 0]
    set align_addr [expr ($addr_lo + 0x3f) & ~0x3f]
    hmem copy $align_addr $pkt_addr0 $pkt_sz
    set pkt_addr0 $align_addr
    set pkt_addr1 [expr $pkt_addr0 + 0x20]

    # Prepare BDs
    set db_idx 0
    set ring_offset [expr $::BDETBD(bde_rcb_2_3_host_phy_adr0) + 8 * $ring]
    set bd_addr [reg read $ring_offset]
    for {set idx 0} {$idx < $pkt_cnt} {incr idx} {
        # Program buffer lengths to be 0x20 for each of the two BDs.
        # First BD...
        if {$pkt_sz < 512} { 
            set lhint 0
        } elseif {$pkt_sz < 1024} { 
            set lhint 1
        } elseif {$pkt_sz < 2048} { 
            set lhint 2
        } else { 
            set lhint 3
        }
        set val [expr ($lhint << 13) | (0x20 << 16) | 0x310]
        hmem write [expr $bd_addr + 0x0] $val
        set opq [expr (($pkt_sz - 4) << 16) | (3 * $idx)]
        hmem write [expr $bd_addr + 0x4] $opq
        hmem write [expr $bd_addr + 0x8] $pkt_addr0
        hmem write [expr $bd_addr + 0xc] 0
        hmem write [expr $bd_addr + 0x10] 0
        hmem write [expr $bd_addr + 0x14] 0
        hmem write [expr $bd_addr + 0x18] 0
        hmem write [expr $bd_addr + 0x1c] 0
        # Second BD...
        set val [expr ($lhint << 13) | (($pkt_sz - 4 - 0x20) << 16) | 0x340]
        hmem write [expr $bd_addr + 0x20] $val 
        hmem write [expr $bd_addr + 0x24] $opq
        hmem write [expr $bd_addr + 0x28] $pkt_addr1
        hmem write [expr $bd_addr + 0x2c] 0
        # One long BD.
        incr bd_addr 0x30
        incr db_idx 3
    }
    return $db_idx
}

proc ring_db {db_idx {ring_idx 0}} {
    # Ring doorbell
    puts "Write [hex32 $db_idx] to DB0"
    bar 1 write [expr 0x80 * $ring_idx] $db_idx 4
}


# Generate a packet
set y [gen_pkt "" 68]
set y120 [gen_pkt "" 120]
set y200 [gen_pkt "" 200]
set y1514 [gen_pkt "" 1514]


